from __future__ import print_function
from __future__ import absolute_import
from builtins import str
from builtins import range
from builtins import object
from six import string_types

# Copyright (c) 2015, Disney Research
# All rights reserved.
#
# Author(s): Sehoon Ha <sehoon.ha@disneyresearch.com>
# Disney Research Robotics Group


import os.path
from .api import pydart2_api as papi
import numpy as np
from .skel_vector import SkelVector

from .bodynode import BodyNode
from .dof import Dof
from .joint import create_joint
from .marker import Marker


class Skeleton(object):
    def __init__(self, _world, _filename=None,
                 _id=None, _friction=None, ):
        self.world = _world
        self.friction = _friction
        if _filename is not None:
            self.filename = os.path.realpath(_filename)
            self.id = papi.world__addSkeleton(self.world.id, self.filename)
        else:
            self.filename = None
            self.id = _id

        self.controller = None
        self.build()

    def build(self, ):
        # Initialize dofs
        _ndofs = papi.skeleton__getNumDofs(self.world.id, self.id)
        self.dofs = [Dof(self, i) for i in range(_ndofs)]
        self.name_to_dof = {dof.name: dof for dof in self.dofs}

        # Initialize joints
        _njoints = papi.skeleton__getNumJoints(self.world.id, self.id)
        self.joints = [create_joint(self, i) for i in range(_njoints)]
        self.name_to_joint = {joint.name: joint for joint in self.joints}

        # Initialize bodynodes
        _nbodynodes = papi.skeleton__getNumBodyNodes(self.world.id, self.id)
        self.bodynodes = [BodyNode(self, i) for i in range(_nbodynodes)]
        self.name_to_body = {body.name: body for body in self.bodynodes}

        # Initialize markers
        _nmarkers = papi.skeleton__getNumMarkers(self.world.id, self.id)
        self.markers = [Marker(self, i) for i in range(_nmarkers)]

        for joint in self.joints:
            joint.build()

        for body in self.bodynodes:
            body.build()

        for marker in self.markers:
            marker.build()

    def set_root_joint_to_trans_and_euler(self, ):
        papi.skeleton__setRootJointToTransAndEuler(self.world.id,
                                                   self.id)
        self.build()

    def set_controller(self, _controller):
        self.controller = _controller

    @property
    def name(self):
        return papi.skeleton__getName(self.world.id, self.id)

    def num_dofs(self):
        return len(self.dofs)

    @property
    def ndofs(self):
        return self.num_dofs()

    def num_joints(self):
        return len(self.joints)

    @property
    def njoints(self):
        return self.num_joints()

    def num_bodynodes(self):
        return len(self.bodynodes)

    @property
    def nbodies(self):
        return self.num_bodynodes()

    def is_mobile(self):
        return papi.skeleton__isMobile(self.world.id, self.id)

    def set_mobile(self, mobile):
        papi.skeleton__setMobile(self.world.id, self.id, mobile)

    def mass(self):
        return papi.skeleton__getMass(self.world.id, self.id)

    @property
    def m(self):
        return self.mass()

    def positions(self):
        q = papi.skeleton__getPositions(self.world.id, self.id, self.ndofs)
        # return q
        return SkelVector(q, self)

    @property
    def q(self):
        return self.positions()

    def set_positions(self, _q):
        papi.skeleton__setPositions(self.world.id, self.id, _q)

    @q.setter
    def q(self, _q):
        """ Setter also updates the internal skeleton kinematics """
        self.set_positions(_q)

    def velocities(self):
        qdot = papi.skeleton__getVelocities(self.world.id, self.id, self.ndofs)
        return SkelVector(qdot, self)

    @property
    def dq(self):
        return self.velocities()

    def set_velocities(self, _qdot):
        papi.skeleton__setVelocities(self.world.id, self.id, _qdot)

    @dq.setter
    def dq(self, _qdot):
        """ Setter also updates the internal skeleton kinematics """
        self.set_velocities(_qdot)

    def accelerations(self):
        ddq = papi.skeleton__getAccelerations(self.world.id,
                                              self.id,
                                              self.ndofs)
        return SkelVector(ddq, self)

    @property
    def ddq(self):
        return self.accelerations()

    def states(self):
        return np.concatenate((self.positions(), self.velocities()))

    @property
    def x(self):
        return np.concatenate((self.positions(), self.velocities()))

    def set_states(self, _x):
        self.set_positions(_x[:self.ndofs])
        self.set_velocities(_x[self.ndofs:])

    @x.setter
    def x(self, _x):
        self.set_states(_x)

    def position_differences(self, q1, q2):
        ret = papi.skeleton__getPositionDifferences(self.world.id, self.id,
                                                    q1, q2, self.ndofs)
        return ret

    def velocity_differences(self, dq1, dq2):
        ret = papi.skeleton__getVelocityDifferences(self.world.id, self.id,
                                                    dq1, dq2, self.ndofs)
        return ret

    def mass_matrix(self):
        M = np.zeros((self.ndofs, self.ndofs))
        papi.skeleton__getMassMatrix(self.world.id, self.id, M)
        return M

    @property
    def M(self):
        return self.mass_matrix()

    def coriolis_and_gravity_forces(self):
        return papi.skeleton__getCoriolisAndGravityForces(self.world.id,
                                                          self.id, self.ndofs)

    @property
    def c(self):
        return self.coriolis_and_gravity_forces()

    def self_collision_check(self):
        return papi.skeleton__getSelfCollisionCheck(self.world.id, self.id)

    def set_self_collision_check(self, _enable):
        papi.skeleton__setSelfCollisionCheck(self.world.id, self.id,
                                             _enable)

    def adjacent_body_check(self):
        return papi.skeleton__getAdjacentBodyCheck(self.world.id, self.id)

    def set_adjacent_body_check(self, _enable):
        papi.skeleton__setAdjacentBodyCheck(self.world.id, self.id,
                                            _enable)

    # def remove_all_collision_pairs(self):
    #     for b1 in self.bodies:
    #         for b2 in self.bodies:
    #             self.world.set_collision_pair(b1, b2, False)

    def constraint_forces(self):
        return papi.skeleton__getConstraintForces(self.world.id,
                                                  self.id, self.ndofs)

    def bodynode(self, query):
        if isinstance(query, string_types):
            return self.name_to_body[query]
        elif isinstance(query, int):
            return self.bodies[query]
        else:
            print('Cannot find body. query = %s' % str(query))
            return None

    def root_bodynodes(self, ):
        return [body for body in self.bodynodes
                if body.parent_bodynode is None]

    def root_bodynode(self, index=0):
        return self.root_bodynodes()[index]

    def bodynode_index(self, _name):
        return self.name_to_body[_name].id

    def body(self, query):
        if isinstance(query, string_types):
            return self.name_to_body[query]
        elif isinstance(query, int):
            return self.bodies[query]
        else:
            print('Cannot find body. query = %s' % str(query))
            return None

    def joint(self, query):
        if isinstance(query, string_types):
            return self.name_to_joint[query]
        elif isinstance(query, int):
            return self.joints[query]
        else:
            print('Cannot find joint. query = %s' % str(query))
            return None

    def dof(self, query):
        if isinstance(query, string_types):
            return self.name_to_dof[query]
        elif isinstance(query, int):
            return self.dofs[query]
        else:
            print('Cannot find dof. query = %s' % str(query))
            return None

    def dof_index(self, _name):
        return self.name_to_dof[_name].id

    def dof_indices(self, _names):
        return np.array([self.dof_index(n) for n in _names])

    def com(self):
        return papi.skeleton__getCOM(self.world.id, self.id)

    @property
    def C(self):
        return self.com()

    def com_velocity(self):
        return papi.skeleton__getCOMLinearVelocity(self.world.id, self.id)

    @property
    def dC(self):
        return self.com_velocity()

    def com_acceleration(self):
        return papi.skeleton__getCOMLinearAcceleration(self.world.id, self.id)

    @property
    def ddC(self):
        return self.com_acceleration()

    def linear_momentum(self):
        return self.Cdot * self.m

    @property
    def P(self):
        return self.linear_momentum()

    def forces(self):
        return self._tau

    @property
    def tau(self):
        return self.forces()

    def set_forces(self, _tau):
        self._tau = _tau
        papi.skeleton__setForces(self.world.id, self.id, _tau)

    @tau.setter
    def tau(self, _tau):
        self.set_forces(_tau)

    def position_lower_limits(self):
        return papi.skeleton__getPositionLowerLimits(self.world.id,
                                                     self.id, self.ndofs)

    def position_upper_limits(self):
        return papi.skeleton__getPositionUpperLimits(self.world.id,
                                                     self.id, self.ndofs)

    @property
    def q_lower(self):
        return self.position_lower_limits()

    @property
    def q_upper(self):
        return self.position_upper_limits()

    def force_lower_limits(self):
        return papi.skeleton__getForceLowerLimits(self.world.id,
                                                  self.id, self.ndofs)

    def force_upper_limits(self):
        return papi.skeleton__getForceUpperLimits(self.world.id,
                                                  self.id, self.ndofs)

    @property
    def tau_lower(self):
        return self.force_lower_limits()

    @property
    def tau_upper(self):
        return self.force_upper_limits()

    # def approx_inertia(self, axis):
    #     """Calculates the point-masses approximated inertia
    #     with respect to the given axis """
    #     axis = np.array(axis) / np.linalg.norm(axis)
    #     I = 0
    #     C = self.C
    #     for body in self.bodies:
    #         d = body.C - C
    #         # Subtract the distance along the axis
    #         r_sq = np.linalg.norm(d) ** 2 - np.linalg.norm(d.dot(axis)) ** 2
    #         I += body.m * r_sq
    #     return I

    # def approx_inertia_x(self):
    #     return self.approx_inertia([1, 0, 0])

    # def approx_inertia_y(self):
    #     return self.approx_inertia([0, 1, 0])

    # def approx_inertia_z(self):
    #     return self.approx_inertia([0, 0, 1])

    # def external_contacts_and_body_id(self):
    #     cid_cnt = dict()
    #     contacts = []
    #     for body in self.bodies:
    #         for c in body.contacts():
    #             contacts += [(c, body.id)]
    #             cid = int(c[6])
    #             if cid not in cid_cnt:
    #                 cid_cnt[cid] = 1
    #             else:
    #                 cid_cnt[cid] += 1
    #     return [(c, bid) for (c, bid) in contacts if cid_cnt[int(c[6])] < 2]

    # def contact_id_set(self):
    #     id_list = []
    #     for b in self.bodies:
    #         id_list += [c.i for c in b.contacts()]
    #     return set(id_list)

    # def contacted_bodies(self):
    #     return [body for body in self.bodies if body.num_contacts() > 0]

    # def world_cop(self):
    #     bodies = self.contacted_bodies()
    #     if len(bodies) == 0:
    #         return None
    #     pos_list = [b.C for b in bodies]
    #     avg = sum(pos_list) / len(pos_list)
    #     return avg

    # @property
    # def COP(self):
    #     return self.world_cop()

    # def contacted_body_names(self):
    #     return [body.name for body in self.contacted_bodies()]

    def render(self):
        papi.skeleton__render(self.world.id, self.id)

    def render_with_color(self, color):
        if len(color) == 3:
            color = np.concatenate([color, [1.0]])
        papi.skeleton__renderWithColor(self.world.id, self.id, color)

    # def render_markers(self):
    #     papi.renderSkeletonMarkers(self.world.id, self.id)

    def __repr__(self):
        return '[Skeleton(%d): %s]' % (self.id, self.name)



# Copyright (c) 2015, Disney Research
# All rights reserved.
#
# Author(s): Sehoon Ha <sehoon.ha@disneyresearch.com>
# Disney Research Robotics Group

from __future__ import print_function
from __future__ import absolute_import
from builtins import chr
from builtins import str
import sys
import os
import signal

# import numpy as np
# import math
# import time
# import inspect

import logging

import OpenGL.GLUT
from PyQt4 import QtGui
from PyQt4 import QtCore
from .glwidget import GLWidget
from .trackball import Trackball


def signal_handler(signal, frame):
    print('You pressed Ctrl+C! Bye.')
    sys.exit(0)


signal.signal(signal.SIGINT, signal_handler)


class PydartWindow(QtGui.QMainWindow):
    GLUT_INITED = False

    def __init__(self, sim=None, title=None):
        if not PydartWindow.GLUT_INITED:
            OpenGL.GLUT.glutInit(sys.argv)

        if title is None:
            title = "Simulation Viewer"
        self.app = QtGui.QApplication([title])

        super(PydartWindow, self).__init__()

        # setup_logger()
        self.logger = logging.getLogger(__name__)
        self.init_callbacks()
        self.set_simulation(sim)
        self.logger.info('sim = [%s]' % str(self.sim))

        # Check and create captures directory
        self.capture_dir = 'data/captures'
        if not os.path.isdir('data/captures'):
            os.makedirs('data/captures')

        self.initUI()
        self.initActions()
        self.initToolbar()
        self.initMenu()
        self.init_cameras()

        self.idleTimer = QtCore.QTimer()
        self.idleTimer.timeout.connect(self.idleTimerEvent)
        self.idleTimer.start(0)

        self.renderTimer = QtCore.QTimer()
        self.renderTimer.timeout.connect(self.renderTimerEvent)
        self.renderTimer.start(80)

        self.camera_event(0)

        self.after_reset = True
        self.set_capture_rate(100)

        self.input_keys = list()
        self.topLeft()

    def run_application(self,):
        self.show()
        self.app.exec_()

    def closeEvent(self, event):
        self.safe_call_callback("on_close")
        super(PydartWindow, self).closeEvent(event)

    def topLeft(self):
        frameGm = self.frameGeometry()
        desktop = QtGui.QApplication.desktop()
        screen = desktop.screenNumber(desktop.cursor().pos())
        # centerPoint = desktop.screenGeometry(screen).center()
        # frameGm.moveCenter(centerPoint)
        topLeftPoint = desktop.screenGeometry(screen).topLeft()
        frameGm.moveTopLeft(topLeftPoint)
        self.move(frameGm.topLeft())

    def get_simulation(self):
        return self.sim

    def set_simulation(self, sim):
        self.sim = sim
        if hasattr(self, "glwidget"):
            self.glwidget.sim = sim
        self.read_default_callbacks_from_sim()

    def capture_rate(self,):
        return self._capture_rate

    def set_capture_rate(self, _capture_rate):
        self._capture_rate = _capture_rate

    def initUI(self):
        TOOLBOX_HEIGHT = 30
        STATUS_HEIGHT = 30
        PANEL_WIDTH = 0
        self.setGeometry(0, 0, 1280 + PANEL_WIDTH,
                         720 + TOOLBOX_HEIGHT + STATUS_HEIGHT)
        # self.setWindowTitle('Toolbar')

        self.ui = QtGui.QWidget(self)
        self.ui.setGeometry(0, TOOLBOX_HEIGHT, 1280 + PANEL_WIDTH, 720)

        self.hbox = QtGui.QHBoxLayout()

        self.glwidget = GLWidget()
        self.glwidget.sim = self.sim
        self.glwidget.viewer = self
        self.glwidget.setGeometry(0, 0, 1280, 720)

        self.hbox.addWidget(self.glwidget)
        self.ui.setLayout(self.hbox)

    def add_right_panel(self, panel):
        TOOLBOX_HEIGHT = 30
        STATUS_HEIGHT = 30
        PANEL_WIDTH = panel.width

        self.setGeometry(0, 0, 1280 + PANEL_WIDTH,
                         720 + TOOLBOX_HEIGHT + STATUS_HEIGHT)
        self.ui.setGeometry(0, TOOLBOX_HEIGHT, 1280 + PANEL_WIDTH, 720)
        self.glwidget.setGeometry(0, 0, 1280, 720)
        self.glwidget.resizeGL(1280, 720)

        self.right_panel = panel
        layout = panel.build_layout()
        self.hbox.addLayout(layout)

        self.ui.setLayout(self.hbox)

    def init_cameras(self,):
        self.cameras = list()
        self.add_camera(
            Trackball(
                rot=[-0.152, 0.045, -0.002, 0.987],
                trans=[0.050, 0.210, -2.500]),
            "Camera Y up")
        self.add_camera(
            Trackball(
                rot=[0.535, 0.284, 0.376, 0.701], trans=[0.10, 0.02, -2.770]),
            "Camera Z up")

    def init_callbacks(self,):
        self.callback_events = ["reset", "step", "render_with_ri",
                                "name",
                                "draw_with_ri",
                                "num_frames", "set_frame",
                                "status", "on_close", "on_key_press",
                                "on_mouse_press", "on_mouse_move",
                                "on_mouse_release"]
        self.callbacks = dict()
        for evt in self.callback_events:
            self.callbacks[evt] = None

    def read_default_callbacks_from_sim(self,):
        self.logger.info('read_default_callbacks_from_sim')
        if self.sim is None:
            return

        for evt in self.callback_events:
            if hasattr(self.sim, evt):
                self.callbacks[evt] = getattr(self.sim, evt)
            else:
                self.callbacks[evt] = None

    def callback(self, evt):
        return self.callbacks[evt]

    def has_callback(self, evt):
        return self.callbacks[evt] is not None

    def set_callback(self, evt, func):
        self.callbacks[evt] = func

    def safe_call_callback(self, evt, *args):
        func = self.callbacks[evt]
        if func is None:
            return None
        return func(*args)

    def callbacks_as_string(self,):
        return "\n".join(["\t[%s: %s]" % (str(evt), str(func))
                          for evt, func in self.callbacks.items()])

    def num_cameras(self,):
        return len(self.cameras)

    def replace_camera(self, idx, trackball):
        if idx >= self.num_cameras():
            return False
        self.cameras[idx] = trackball
        return True

    def add_camera(self, trackball, name=None):
        self.cameras.append(trackball)
        cam_id = len(self.cameras) - 1
        if name is None:
            name = 'Camera %d' % cam_id
        action = QtGui.QAction(name, self)
        action.triggered.connect(lambda: self.camera_event(cam_id))
        self.cameraMenu.addAction(action)

    def initActions(self):
        # Create actions
        self.resetAction = QtGui.QAction('Reset', self)
        self.resetAction.triggered.connect(self.resetEvent)

        self.stepAction = QtGui.QAction('Step', self)
        self.stepAction.triggered.connect(self.stepEvent)

        self.playAction = QtGui.QAction('Play', self)
        self.playAction.setCheckable(True)
        self.playAction.setShortcut('Space')

        self.animAction = QtGui.QAction('Anim', self)
        self.animAction.setCheckable(True)

        self.captureAction = QtGui.QAction('Capture', self)
        self.captureAction.setCheckable(True)

        self.emptyAction = QtGui.QAction('Empty', self)
        self.emptyAction.triggered.connect(self.emptyEvent)

        self.movieAction = QtGui.QAction('Movie', self)
        self.movieAction.triggered.connect(self.movieEvent)

        self.screenshotAction = QtGui.QAction('Screenshot', self)
        self.screenshotAction.triggered.connect(self.screenshotEvent)

        self.printCamAction = QtGui.QAction('Print Camera', self)
        self.printCamAction.triggered.connect(self.printCamEvent)

    def initToolbar(self):
        # Create a toolbar
        self.toolbar = self.addToolBar('Control')
        self.toolbar.addAction(self.resetAction)
        self.toolbar.addAction(self.stepAction)
        self.toolbar.addAction(self.playAction)
        self.toolbar.addAction(self.animAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.screenshotAction)
        self.toolbar.addAction(self.captureAction)
        self.toolbar.addAction(self.emptyAction)
        self.toolbar.addAction(self.movieAction)

        self.rangeSlider = QtGui.QSlider(QtCore.Qt.Horizontal, self)
        self.rangeSlider.valueChanged[int].connect(self.rangeSliderEvent)
        self.toolbar.addWidget(self.rangeSlider)

    def initMenu(self):
        menubar = self.menuBar()
        fileMenu = menubar.addMenu('&File')
        fileMenu.addSeparator()

        # Camera menu
        self.cameraMenu = menubar.addMenu('&Camera')
        self.cameraMenu.addAction(self.printCamAction)
        self.cameraMenu.addSeparator()

        # Recording menu
        recordingMenu = menubar.addMenu('&Recording')
        recordingMenu.addAction(self.screenshotAction)
        recordingMenu.addSeparator()
        recordingMenu.addAction(self.captureAction)
        recordingMenu.addAction(self.movieAction)

    def idleTimerEvent(self):
        doCapture = False

        # Do animation
        if self.animAction.isChecked():
            v = self.rangeSlider.value()
            v += 1
            if v <= self.rangeSlider.maximum():
                self.rangeSlider.setValue(v)
            else:
                self.animAction.setChecked(False)
            doCapture = True

        # Do play
        elif self.playAction.isChecked():
            bIsDone = self.safe_call_callback('step')
            if bIsDone is False:
                print("step() returns False: stop...")
                self.playAction.setChecked(False)
            # capture_rate = 10
            doCapture = (self.sim.frame % self.capture_rate() == 0)
            # doCapture = True

        if self.captureAction.isChecked() and doCapture:
            self.glwidget.capture(self.prob_name())

    def renderTimerEvent(self):
        self.glwidget.updateGL()

        msg = self.safe_call_callback('status')
        if msg is None:
            msg = str(self.sim)
        self.statusBar().showMessage(msg)

        n = self.safe_call_callback('num_frames')
        if n is not None:
            self.rangeSlider.setRange(0, n - 1)

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Escape:
            print('Escape key pressed! Bye.')
            self.close()
            return
        if 0 <= event.key() and event.key() < 256:  # If key is ascii
            key = chr(event.key())
            self.safe_call_callback("on_key_press", key)

    def rangeSliderEvent(self, frame):
        self.safe_call_callback("set_frame", frame)

    def screenshotEvent(self):
        self.glwidget.capture("%s_frame" % self.prob_name())

    def prob_name(self):
        name = self.safe_call_callback("name")
        return name if name is not None else "robot"

    def emptyEvent(self):
        cmd = 'rm %s/*.png' % self.capture_dir
        print('cmd = %s' % cmd)
        os.system(cmd)

    def movieEvent(self):
        """
        Ubuntu tested only: avconv is required
        """
        name = self.prob_name()
        # yuv420p for compatibility for outdated codecs
        cmt_fmt = 'avconv -r 100'
        cmt_fmt += ' -i %s/%s.%%04d.png'
        cmt_fmt += ' -pix_fmt yuv420p'
        cmt_fmt += ' %s.mp4'
        print(self.capture_dir)
        cmd = cmt_fmt % (self.capture_dir, name, name)
        print('cmd = %s' % cmd)
        os.system(cmd)

    def resetEvent(self):
        self.after_reset = True
        self.rangeSlider.setValue(0)
        # if hasattr(self.sim, 'reset'):
        #     self.sim.reset()
        # if self.callbacks['reset'] is not None:
        #     self.callbacks['reset']()
        self.safe_call_callback('reset')

    def stepEvent(self):
        self.safe_call_callback("step")

    def camera_event(self, cam_id):
        print("camera_event: %d" % cam_id)
        self.glwidget.tb = self.cameras[cam_id]

    def printCamEvent(self):
        print('printCamEvent')
        print('----')
        print(repr(self.glwidget.tb))
        print('----')


def launch(sim, title=None, default_camera=None):
    # glutInit(sys.argv)
    win = PydartWindow(sim, title)
    if default_camera is not None:
        win.camera_event(default_camera)
    win.run_application()



import sys
import logging


def setup(filename=None):
    logging.basicConfig()

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.handlers = []
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    logfmt = '[%(levelname)s][%(asctime)s][%(name)s:%(lineno)d] %(message)s'
    formatter = logging.Formatter(logfmt)
    ch.setFormatter(formatter)
    root.addHandler(ch)

    # set up logging to console
    if filename is not None:
        # filename = "output.log"
        console = logging.FileHandler(filename, "w+")
        console.setLevel(logging.DEBUG)
        console.setFormatter(formatter)
        root.addHandler(console)

    logging.addLevelName(logging.FATAL,
                         "\033[1;41m%s\033[1;0m" %
                         logging.getLevelName(logging.FATAL))
    logging.addLevelName(logging.ERROR,
                         "\033[1;41m%s\033[1;0m" %
                         logging.getLevelName(logging.ERROR))
    logging.addLevelName(logging.WARNING,
                         "\033[1;43m%s\033[1;0m" %
                         logging.getLevelName(logging.WARNING))
    logging.addLevelName(logging.INFO,
                         "\033[1;42m%s\033[1;0m" %
                         logging.getLevelName(logging.INFO))
    logging.addLevelName(logging.DEBUG,
                         "\033[1;44m%s\033[1;0m" %
                         logging.getLevelName(logging.DEBUG))



from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from past.utils import old_div
# Copyright (c) 2015, Disney Research
# All rights reserved.
#
# Author(s): Sehoon Ha <sehoon.ha@disneyresearch.com>
# Disney Research Robotics Group
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
from PyQt4 import QtGui
from PyQt4 import QtCore
from PyQt4.QtOpenGL import *
from . import trackball
import time
from .renderer import Renderer
import numpy as np
from numpy.linalg import norm


class GLWidget(QGLWidget):

    def __init__(self, parent=None):
        super(GLWidget, self).__init__(parent)
        fmt = QGLFormat()
        fmt.setSampleBuffers(True)
        fmt.setSamples(4)
        # print fmt.samples()
        # exit(0)

        self.setFormat(fmt)  # Anti-aliasing
        self.width = 1280
        self.height = 720

        self.tb = trackball.Trackball()
        self.lastPos = None
        self.zoom = -1.2

        self.sim = None
        self.viewer = None
        self.captureIndex = 0
        self.renderer = Renderer()

        self.lock_camera = False

    def sizeHint(self):
        return QtCore.QSize(self.width, self.height)

    def paintGL(self):
        glEnable(GL_DEPTH_TEST)
        glClearColor(0.98, 0.98, 0.98, 0.0)
        glClearColor(1.0, 1.0, 1.0, 1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        glLoadIdentity()
        # glTranslate(0.0, -0.2, self.zoom)  # Camera
        glTranslate(*self.tb.trans)
        glMultMatrixf(self.tb.matrix)

        if self.sim is not None and hasattr(self.sim, "render"):
            self.sim.render()

        self.renderer.enable("COLOR_MATERIAL")
        if self.sim is not None and hasattr(self.sim, "render_with_ri"):
            self.sim.render_with_ri(self.renderer)

        self.enable2D()
        if self.sim is not None and hasattr(self.sim, "draw_with_ri"):
            self.sim.draw_with_ri(self.renderer)
            self.renderer.draw_text([-100, -100], "")
        self.disable2D()

    def resizeGL(self, w, h):
        (self.width, self.height) = (w, h)
        glViewport(0, 0, w, h)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        gluPerspective(45.0, old_div(float(w), float(h)), 0.01, 100.0)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

    def initializeGL(self):
        # return self.initializeGL_Stelian()
        glDisable(GL_CULL_FACE)
        glEnable(GL_DEPTH_TEST)

        glDepthFunc(GL_LEQUAL)
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST)

        glEnable(GL_LINE_SMOOTH)
        glHint(GL_LINE_SMOOTH_HINT, GL_NICEST)
        # GlEnable(GL_POLYGON_SMOOTH)
        glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST)

        glEnable(GL_DITHER)
        glShadeModel(GL_SMOOTH)
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST)

        glClearColor(1.0, 1.0, 1.0, 1.0)
        glClear(GL_COLOR_BUFFER_BIT)

        glEnable(GL_DEPTH_TEST)
        glDepthFunc(GL_LEQUAL)
        glDisable(GL_CULL_FACE)
        glEnable(GL_NORMALIZE)

        glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE)
        glEnable(GL_COLOR_MATERIAL)

        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glEnable(GL_MULTISAMPLE)
        # glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA)

        ambient = [0.2, 0.2, 0.2, 1.0]
        diffuse = [0.6, 0.6, 0.6, 1.0]
        front_mat_shininess = [60.0]
        front_mat_specular = [0.2, 0.2, 0.2, 1.0]
        front_mat_diffuse = [0.5, 0.28, 0.38, 1.0]
        lmodel_ambient = [0.2, 0.2, 0.2, 1.0]
        lmodel_twoside = [GL_FALSE]

        # position = [1.0, 1.0, 1.0, 0.0]
        # position1 = [-1.0, 1.0, 0.0, 0.0]

        position = [1.0, 1.0, 0.0, 0.0]
        position1 = [-1.0, 0.0, 0.0, 0.0]

        glEnable(GL_LIGHT0)
        glLightfv(GL_LIGHT0, GL_AMBIENT, ambient)
        glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse)
        glLightfv(GL_LIGHT0, GL_POSITION, position)

        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient)
        glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lmodel_twoside)

        glEnable(GL_LIGHT1)
        # glLightfv(GL_LIGHT1, GL_AMBIENT, ambient)
        glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse)
        glLightfv(GL_LIGHT1, GL_POSITION, position1)
        glEnable(GL_LIGHTING)

        glEnable(GL_COLOR_MATERIAL)
        glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, front_mat_shininess)
        glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, front_mat_specular)
        glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, front_mat_diffuse)

        # glBlendFunc(GL_ONE, GL_ONE)
        # glCullFace(GL_BACK)

    def enable2D(self):
        w, h = self.width, self.height
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glDisable(GL_LIGHTING | GL_DEPTH_TEST)
        glDepthMask(0)
        glOrtho(0, w, h, 0, -1, 1)
        glViewport(0, 0, w, h)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

    def disable2D(self):
        w, h = self.width, self.height
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        glEnable(GL_DEPTH_TEST | GL_LIGHTING)
        glDepthMask(1)
        gluPerspective(45.0, old_div(float(w), float(h)), 0.01, 100.0)

        glViewport(0, 0, w, h)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

    def set_lock_camera(self, lock=True):
        self.lock_camera = lock

    def mousePressEvent(self, event):
        self.lastPos = event.pos()

        pos = np.array([event.pos().x(), event.pos().y()],
                       dtype=np.float64)
        self.viewer.safe_call_callback("on_mouse_press", pos)

    def mouseReleaseEvent(self, event):
        self.lastPos = None
        self.viewer.safe_call_callback("on_mouse_release")

    def mouseMoveEvent(self, event):
        # (w, h) = (self.width, self.height)
        x = event.x()
        y = event.y()
        dx = event.x() - self.lastPos.x()
        dy = event.y() - self.lastPos.y()

        if not self.lock_camera:
            modifiers = QtGui.QApplication.keyboardModifiers()
            if modifiers == QtCore.Qt.ShiftModifier:
                self.tb.zoom_to(dx, -dy)
            elif modifiers == QtCore.Qt.ControlModifier:
                self.tb.trans_to(dx, -dy)
            else:
                self.tb.drag_to(x, y, dx, -dy)

        p0 = np.array([self.lastPos.x(), self.lastPos.y()],
                      dtype=np.float64)
        p1 = np.array([event.pos().x(), event.pos().y()],
                      dtype=np.float64)
        self.viewer.safe_call_callback("on_mouse_move", p0, p1)

        self.lastPos = event.pos()
        self.updateGL()

    def capture(self, name=None):
        img = self.grabFrameBuffer()
        if name is None:
            name = 'frame'
        dir = self.viewer.capture_dir
        filename = '%s/%s.%04d.png' % (dir, name, self.captureIndex)
        img.save(filename)
        print(('Capture to %s' % filename))
        self.captureIndex += 1



from __future__ import absolute_import
from builtins import object
# Copyright (c) 2015, Disney Research
# All rights reserved.
#
# Author(s): Sehoon Ha <sehoon.ha@disneyresearch.com>
# Disney Research Robotics Group
from .api import pydart2_api as papi


class Shape(object):
    """
    """
    def __init__(self, _shapenode):
        """
        """
        self.shapenode = _shapenode
        self.bodynode = _shapenode.bodynode
        self.skeleton = _shapenode.bodynode.skeleton
        self.id = _shapenode.id

    @property
    def skel(self):
        return self.skeleton

    @property
    def wid(self):
        return self.skel.world.id

    @property
    def skid(self):
        return self.skel.id

    @property
    def bid(self):
        return self.bodynode.id

    def build(self):
        pass

    def volume(self, ):
        return papi.shape__getVolume(self.wid, self.skid, self.bid, self.id)

    def shape_type(self, ):
        type_ = papi.shape__getShapeType(self.wid, self.skid,
                                         self.bid, self.id)
        type_ = 1 if type_ == -1 else type_  # Assume to be ELLIPSOID
        return type_

    def shape_type_name(self, ):
        type_ = self.shape_type()
        names = ["BOX", "ELLIPSOID", "CYLINDER", "PLANE",
                 "MESH", "SOFT_MESH", "LINE_SEGMENT"]
        return names[type_]

    def render(self, ):
        papi.shape__render(self.wid, self.skid, self.bid, self.id)

    def bounding_box(self, ):
        _min = papi.shape__getBoundingBoxMin(self.wid,
                                             self.skid,
                                             self.bid,
                                             self.id)
        _max = papi.shape__getBoundingBoxMax(self.wid,
                                             self.skid,
                                             self.bid,
                                             self.id)
        return (_min, _max)

    def __repr__(self):
        return '[Shape(%d:%d)]' % (self.bid, self.id)



from __future__ import absolute_import
from builtins import object
# Copyright (c) 2015, Disney Research
# All rights reserved.
#
# Author(s): Sehoon Ha <sehoon.ha@disneyresearch.com>
# Disney Research Robotics Group
from .api import pydart2_api as papi


class BallJointConstraint(object):
    def __init__(self, body1, body2, jointPos):
        self.body1 = body1
        self.body2 = body2
        self.jointPos = jointPos

    def add_to_world(self, world):
        papi.addBallJointConstraint(world.id,
                                    self.body1.skid,
                                    self.body1.id,
                                    self.body2.skid,
                                    self.body2.id,
                                    self.jointPos)



from __future__ import absolute_import
from builtins import object
# Copyright (c) 2015, Disney Research
# All rights reserved.
#
# Author(s): Sehoon Ha <sehoon.ha@disneyresearch.com>
# Disney Research Robotics Group
from .api import pydart2_api as papi


class Marker(object):
    """
    """
    def __init__(self, _skeleton, _id):
        """
        """
        self.skeleton = _skeleton
        self.id = _id
        self.name = "M%04d" % _id
        self.bodynode = None

    @property
    def skel(self):
        return self.skeleton

    @property
    def wid(self):
        return self.skel.world.id

    @property
    def skid(self):
        return self.skel.id

    def build(self):
        self.bodynode = None

        ret_id = papi.marker__getBodyNode(self.wid, self.skid, self.id)
        if ret_id >= 0:
            self.bodynode = self.skel.bodynodes[ret_id]
            self.bodynode.markers.append(self)

    def local_position(self, ):
        return papi.marker__getLocalPosition(self.wid, self.skid, self.id)

    def set_local_position(self, inv3):
        papi.marker__setLocalPosition(self.wid, self.skid, self.id, inv3)

    def world_position(self, ):
        return papi.marker__getWorldPosition(self.wid, self.skid, self.id)

    @property
    def p(self, ):
        return self.world_position

    def render(self, ):
        return papi.marker__render(self.wid, self.skid, self.id)

    def __repr__(self):
        return '[Marker(%d)]' % (self.id)



import unittest
import pydart2 as pydart
from tests.config import DATA_DIR


class TestBasic(unittest.TestCase):
    def setUp(self):
        pydart.init(verbose=False)

    def test_world_create(self):
        world = pydart.world.World(0.001)
        self.assertIsNotNone(world)
        self.assertEqual(world.id, 0)

    def test_world_from_skeleton(self, ):
        world = pydart.world.World(
            0.001,
            DATA_DIR + "skel/cubes.skel"
            )
        self.assertIsNotNone(world)
        self.assertEqual(world.id, 0)

    def test_cube_collisions(self, ):
        world = pydart.world.World(
            0.001,
            DATA_DIR + "skel/cubes.skel"
            )
        self.assertIsNotNone(world)
        self.assertEqual(world.id, 0)

        skel = world.skeletons[-1]
        while world.t < 3.0:
            world.step()
        # target = [7.94625426e-02, -3.25000010e-01, -9.22240587e-18]
        # self.assertAlmostEqual(skel.C[0], target[0])
        # self.assertAlmostEqual(skel.C[1], target[1])
        # self.assertAlmostEqual(skel.C[2], target[2])
        self.assertLess(0.0, skel.C[0])



from __future__ import absolute_import
from builtins import range
from builtins import object
# Copyright (c) 2015, Disney Research
# All rights reserved.
#
# Author(s): Sehoon Ha <sehoon.ha@disneyresearch.com>
# Disney Research Robotics Group
from .api import pydart2_api as papi
import numpy as np
from .shapenode import ShapeNode


class BodyNode(object):
    def __init__(self, _skel, _id):
        self.skeleton = _skel
        self._id = _id
        self.name = papi.bodynode__getName(self.wid, self.skid, self.id)
        self.parent_bodynode = None
        self.child_bodynodes = list()

        self.parent_joint = None
        self.child_joints = list()

        self.dependent_dofs = list()
        self.markers = list()  # Built by markers

        self.shapenodes = list()

    def build(self):
        # Build Body Nodes
        self.parent_bodynode = None
        self.child_bodynodes = list()

        ret_id = papi.bodynode__getParentBodyNode(self.wid, self.skid, self.id)
        if ret_id >= 0:
            self.parent_bodynode = self.skel.bodynodes[ret_id]

        num = papi.bodynode__getNumChildBodyNodes(self.wid, self.skid, self.id)
        for index in range(num):
            ret_id = papi.bodynode__getChildBodyNode(self.wid,
                                                     self.skid,
                                                     self.id,
                                                     index)
            if ret_id >= 0:
                self.child_bodynodes.append(self.skel.bodynodes[ret_id])

        # Build Joints
        self.parent_joint = None
        self.child_joints = list()
        ret_id = papi.bodynode__getParentJoint(self.wid, self.skid, self.id)
        if ret_id >= 0:
            self.parent_joint = self.skel.joints[ret_id]

        num = papi.bodynode__getNumChildJoints(self.wid, self.skid, self.id)
        for index in range(num):
            ret_id = papi.bodynode__getChildJoint(self.wid,
                                                  self.skid,
                                                  self.id,
                                                  index)
            if ret_id >= 0:
                self.child_joints.append(self.skel.joints[ret_id])

        # Build dependent_dofs
        self.dependent_dofs = list()
        num = papi.bodynode__getNumDependentDofs(self.wid, self.skid, self.id)
        for index in range(num):
            ret_id = papi.bodynode__getDependentDof(self.wid,
                                                    self.skid,
                                                    self.id,
                                                    index)
            if ret_id >= 0:
                self.dependent_dofs.append(self.skel.dofs[ret_id])

        # Build shapenodes
        self.shapenodes = list()
        num = papi.bodynode__getNumShapeNodes(self.wid, self.skid, self.id)
        for index in range(num):
            self.shapenodes.append(ShapeNode(self, index))

        # Build markers (by other class)
        self.markers = list()

    def num_child_bodynodes(self, ):
        return len(self.child_bodynodes)

    def num_dependent_dofs(self, ):
        return len(self.dependent_dofs)

    def num_child_joints(self, ):
        return len(self.child_joints)

    def num_markers(self, ):
        return len(self.markers)

    def num_shapenodes(self, ):
        return len(self.shapenodes)

    @property
    def id(self):
        return self._id

    @property
    def wid(self):
        return self.skel.world.id

    @property
    def skel(self):
        return self.skeleton

    @property
    def skid(self):
        return self.skel.id

    # def num_contacts(self):
    #     return papi.getBodyNodeNumContacts(self.wid, self.skid, self.id)

    # def contacts(self):
    #     n = self.num_contacts()
    #     contacts = papi.getBodyNodeContacts(self.wid,
    #                                         self.skid,
    #                                         self.id, 7 * n)
    #     return [Contact(contacts[7 * i: 7 * (i + 1)]) for i in range(n)]

########################################
# Index Functions
    def index_in_skeleton(self, ):
        return papi.bodynode__getIndexInSkeleton(self.wid, self.skid, self.id)

    def index_in_tree(self, ):
        return papi.bodynode__getIndexInTree(self.wid, self.skid, self.id)

    def tree_index(self, ):
        return papi.bodynode__getTreeIndex(self.wid, self.skid, self.id)

########################################
# Property Functions
    def set_gravity_mode(self, _gravityMode):
        papi.bodynode__setGravityMode(self.wid,
                                      self.skid,
                                      self.id,
                                      _gravityMode)

    def gravity_mode(self, ):
        return papi.bodynode__getGravityMode(self.wid, self.skid, self.id)

    def is_collidable(self, ):
        return papi.bodynode__isCollidable(self.wid, self.skid, self.id)

    def set_collidable(self, _isCollidable):
        papi.bodynode__setCollidable(self.wid,
                                     self.skid,
                                     self.id,
                                     _isCollidable)

########################################
# Inertia Functions
    def mass(self):
        return papi.bodynode__getMass(self.wid, self.skid, self.id)

    @property
    def m(self):
        return self.mass()

    def inertia(self):
        return papi.bodynode__getInertia(self.wid, self.skid, self.id)

    @property
    def I(self):
        return self.inertia()

########################################
# Momentum Functions
    def local_com(self, ):
        return papi.bodynode__getLocalCOM(self.wid, self.skid, self.id)

    def com(self, ):
        return papi.bodynode__getCOM(self.wid, self.skid, self.id)

    @property
    def C(self):
        return self.com()

    def com_linear_velocity(self, ):
        return papi.bodynode__getCOMLinearVelocity(self.wid,
                                                   self.skid,
                                                   self.id)

    @property
    def dC(self):
        return self.com_linear_velocity()

    def com_spatial_velocity(self, ):
        return papi.bodynode__getCOMSpatialVelocity(self.wid,
                                                    self.skid,
                                                    self.id)

    def com_linear_acceleration(self, ):
        return papi.bodynode__getCOMLinearAcceleration(self.wid,
                                                       self.skid,
                                                       self.id)

    def com_spatial_acceleration(self, ):
        return papi.bodynode__getCOMSpatialAcceleration(self.wid,
                                                        self.skid,
                                                        self.id)

########################################
# Friction and Restitution Functions
    def set_friction_coeff(self, _coeff):
        papi.bodynode__setFrictionCoeff(self.wid, self.skid, self.id, _coeff)

    def friction_coeff(self, ):
        return papi.bodynode__getFrictionCoeff(self.wid, self.skid, self.id)

    def set_restitution_coeff(self, _coeff):
        papi.bodynode__setRestitutionCoeff(self.wid,
                                           self.skid,
                                           self.id,
                                           _coeff)

    def restitution_coeff(self, ):
        return papi.bodynode__getRestitutionCoeff(self.wid, self.skid, self.id)

########################################
# Transform Functions
    def transform(self, ):
        return papi.bodynode__getTransform(self.wid, self.skid, self.id)

    @property
    def T(self):
        return self.transform()

    def to_world(self, x=None):
        if x is None:
            x = [0.0, 0.0, 0.0]
        x_ = np.append(x, [1.0])
        return (self.T.dot(x_))[:3]

    def to_local(self, x):
        Tinv = np.linalg.inv(self.T)
        x_ = np.append(x, [1.0])
        return (Tinv.dot(x_))[:3]

    def world_transform(self, ):
        return papi.bodynode__getWorldTransform(self.wid, self.skid, self.id)

    def relative_transform(self, ):
        return papi.bodynode__getRelativeTransform(self.wid,
                                                   self.skid,
                                                   self.id)

########################################
# Torque Functions
    def add_ext_force(self,
                      _force,
                      _offset=None,
                      _isForceLocal=False,
                      _isOffsetLocal=True):
        if _offset is None:
            _offset = np.zeros(3)
        papi.bodynode__addExtForce(self.wid,
                                   self.skid,
                                   self.id,
                                   _force,
                                   _offset,
                                   _isForceLocal,
                                   _isOffsetLocal)

    def set_ext_force(self,
                      _force,
                      _offset=None,
                      _isForceLocal=False,
                      _isOffsetLocal=True):
        if _offset is None:
            _offset = np.zeros(3)
        papi.bodynode__setExtForce(self.wid,
                                   self.skid,
                                   self.id,
                                   _force,
                                   _offset,
                                   _isForceLocal,
                                   _isOffsetLocal)

    def add_ext_torque(self, _torque, _isLocal=False):
        papi.bodynode__addExtTorque(self.wid, self.skid, self.id,
                                    _torque, _isLocal)

    def set_ext_torque(self, _torque, _isLocal=False):
        papi.bodynode__setExtTorque(self.wid, self.skid, self.id,
                                    _torque, _isLocal)

########################################
# Jacobian Functions
    def jacobian(self, offset=None, full=True):
        offset = np.zeros(3) if offset is None else offset
        J = np.zeros((6, len(self.dependent_dofs)))
        papi.bodynode__getJacobian(self.wid,
                                   self.skid,
                                   self.id,
                                   offset,
                                   J)
        return self.expand_jacobian(J) if full else J

    def linear_jacobian(self, offset=None, full=True):
        offset = np.zeros(3) if offset is None else offset
        J = np.zeros((3, len(self.dependent_dofs)))
        papi.bodynode__getLinearJacobian(self.wid,
                                         self.skid,
                                         self.id,
                                         offset,
                                         J)
        return self.expand_jacobian(J) if full else J

    def angular_jacobian(self, full=True):
        J = np.zeros((3, len(self.dependent_dofs)))
        papi.bodynode__getAngularJacobian(self.wid, self.skid, self.id, J)
        return self.expand_jacobian(J) if full else J

    def world_jacobian(self, offset=None, full=True):
        offset = np.zeros(3) if offset is None else offset
        J = np.zeros((6, len(self.dependent_dofs)))
        papi.bodynode__getWorldJacobian(self.wid,
                                        self.skid,
                                        self.id,
                                        offset,
                                        J)
        return self.expand_jacobian(J) if full else J

    def linear_jacobian_deriv(self, offset=None, full=True):
        offset = np.zeros(3) if offset is None else offset
        J = np.zeros((3, len(self.dependent_dofs)))
        papi.bodynode__getLinearJacobianDeriv(self.wid,
                                              self.skid,
                                              self.id,
                                              offset,
                                              J)
        return self.expand_jacobian(J) if full else J

    def angular_jacobian_deriv(self, full=True):
        J = np.zeros((3, len(self.dependent_dofs)))
        papi.bodynode__getAngularJacobianDeriv(self.wid, self.skid, self.id, J)
        return self.expand_jacobian(J) if full else J

    def expand_jacobian(self, jacobian):
        J = jacobian
        F = np.zeros((J.shape[0], self.skel.ndofs))
        I = np.array([dof.index for dof in self.dependent_dofs])
        F[:, I] = J
        return F

    @property
    def J(self, ):
        return self.linear_jacobian()

    @property
    def dJ(self, ):
        return self.linear_jacobian_deriv()

    def __repr__(self):
        return '[BodyNode(%d): %s]' % (self.id, self.name)
